#include "population.h"



Population::Population(int count,QStringList b,QStringList u,QStringList v,QStringList con,double (*ff)(Node *))
{
    chromosome.resize(count);
    children.resize(count);
    binarySet=b;
    unarySet=u;
    variableSet=v;
    constantSet=con;
    fitness_function=ff;
    selection_rate=0.1;
    mutation_rate=0.05;
    fitnessArray.resize(count);
    //initialize chromosomes
    const int maxDepth=5;
    for(int i=0;i<count;i++)
    {
        chromosome[i]=makeNode(NULL,maxDepth);
        children[i]=new Node(NULL,"",0);
      //  printf("c[%d]=%s\n",i,chromosome[i]->print().toStdString().c_str());
    }
    calcFitnessArray();
}

Node    *Population::makeNode(Node *rootNode,int maxDepth)
{
    Node *n;
    int type;
    if(maxDepth<=0)
    {
        int r=rand() %2;
        if(r==0)
        {
            type=NODE_VARIABLE;
            n=new Node(rootNode,variableSet[rand() % variableSet.size()],type);
        }
        else
        {
            type=NODE_CONSTANT;
            n=new Node(rootNode,constantSet[rand() % constantSet.size()],type);
        }
    }
    else
    {
        int r=rand() %2;
        if(r==0)
        {
            type=NODE_UNARY;
            n=new Node(rootNode,unarySet[rand() % unarySet.size()],type);
            Node *n1=makeNode(n,maxDepth-1);
            n->setleft(n1);
        }
        else
        {
             type=NODE_BINARY;
             n=new Node(rootNode,binarySet[rand() % binarySet.size()],type);
             Node *n1=makeNode(n,maxDepth-1);
             n->setleft(n1);
             Node *n2=makeNode(n,maxDepth-1);
             n->setright(n2);
        }
    }
    return n;
}

void    Population::setSelectionRate(double r)
{
    if(r>=0 && r<=1) selection_rate=r;
}

double  Population::getSelectionRate()
{
    return selection_rate;
}

void    Population::setMutationRate(double r)
{
    if(r>=0 && r<=1) mutation_rate=r;
}

double  Population::getMutationRate()
{
    return mutation_rate;
}

void    Population::sortChromosomes()
{
    for(int i=0;i<chromosome.size();i++)
    {
        for(int j=0;j<chromosome.size()-1;j++)
        {
            if(fitnessArray[j+1]<fitnessArray[j])
            {
                double f=fitnessArray[j];
                fitnessArray[j]=fitnessArray[j+1];
                fitnessArray[j+1]=f;
                Node *t=chromosome[j];
                chromosome[j]=chromosome[j+1];
                chromosome[j+1]=t;
            }
        }
    }
}

int     Population::tournament()
{
    const int N=8;
    int bestIndex=0;
    double bestF=0;
    for(int i=1;i<=N;i++)
    {
        int pos=rand() % chromosome.size();
        double f=fitnessArray[pos];
        if(i==1 || f<bestF)
        {
            bestIndex=pos;
            bestF=f;
        }
    }
    return bestIndex;
}

void    Population::crossover()
{
    int nchildren=(int)((1.0  - selection_rate)*chromosome.size());

    int count_children=0;
    while(true)
    {
        int p1=tournament();
        int p2=tournament();


        children[count_children]->copyFromOther(chromosome[p1]);
        Node *randomp2node=chromosome[p2]->randomNode(1.0/chromosome[p2]->getdepth());
        if(randomp2node==NULL) continue;
        children[count_children++]->setRandomNode(1.0/chromosome[p1]->getdepth(),randomp2node);
        if(count_children>=nchildren) break;
    }
    for(int i=0;i<nchildren;i++)
    {
        chromosome[chromosome.size()-i-1]->copyFromOther(children[i]);
    }
}

void    Population::mutate()
{
    for(int i=1;i<chromosome.size();i++)
    {
        const int iters=10;
        for(int i=1;i<=iters;i++)
        {
            double r=rand() *1.0/RAND_MAX;
            if(r<=mutation_rate)
            {
                Node *p=makeNode(NULL,5);
                chromosome[i]->setRandomNode(1.0/chromosome[i]->getdepth(),p);
            }
        }
    }
}

Node    *Population::getBest()
{
    return chromosome[0];
}

double  Population::getBestFitness()
{
    return fitnessArray[0];
}

void    Population::calcFitnessArray()
{
    for(int i=0;i<chromosome.size();i++){

        fitnessArray[i]=fitness_function(chromosome[i]);
    }
}

void    Population::nextGeneration()
{
    ++generation;
    calcFitnessArray();
    sortChromosomes();
    crossover();
    mutate();
}

Population::~Population()
{
    for(int i=0;i<chromosome.size();i++)
        delete chromosome[i];
    for(int j=0;j<children.size();j++)
        delete children[j];
}
